import asyncio
from typing import List
import pytest

from etcetra import EtcdClient, WatchEvent, WatchEventType


@pytest.mark.asyncio
async def test_watch(etcd: EtcdClient):
    events: List[WatchEvent] = []

    async def _put_task():
        await asyncio.sleep(1)
        async with etcd.connect() as communicator:
            await communicator.put('/tmp/asd', 'fgh')
            await asyncio.sleep(0.5)
            await communicator.put('/tmp/qwe', 'rty')
            await asyncio.sleep(0.5)
            await communicator.delete('/tmp/qwe')

    async def _watch_task():
        async with etcd.connect() as communicator:
            async for event in communicator.watch_prefix('/tmp'):
                events.append(event)
        return events

    put_task = asyncio.create_task(_put_task())
    watch_task = asyncio.create_task(_watch_task())

    await asyncio.sleep(3)
    put_task.cancel()
    watch_task.cancel()

    assert events[0] == WatchEvent(
        key='/tmp/asd', value='fgh', prev_value=None, event=WatchEventType.PUT)
    assert events[1] == WatchEvent(
        key='/tmp/qwe', value='rty', prev_value=None, event=WatchEventType.PUT)
    assert events[2] == WatchEvent(
        key='/tmp/qwe', value='', prev_value=None, event=WatchEventType.DELETE)
    assert len(events) == 3



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: raft_internal.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from etcetra.grpc_api import gogo_pb2 as gogo__pb2
from etcetra.grpc_api import etcdserver_pb2 as etcdserver__pb2
from etcetra.grpc_api import rpc_pb2 as rpc__pb2
from etcetra.grpc_api import membership_pb2 as membership__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13raft_internal.proto\x12\x0c\x65tcdserverpb\x1a\ngogo.proto\x1a\x10\x65tcdserver.proto\x1a\trpc.proto\x1a\x10membership.proto\"D\n\rRequestHeader\x12\n\n\x02ID\x18\x01 \x01(\x04\x12\x10\n\x08username\x18\x02 \x01(\t\x12\x15\n\rauth_revision\x18\x03 \x01(\x04\"\xd9\x0e\n\x13InternalRaftRequest\x12+\n\x06header\x18\x64 \x01(\x0b\x32\x1b.etcdserverpb.RequestHeader\x12\n\n\x02ID\x18\x01 \x01(\x04\x12!\n\x02v2\x18\x02 \x01(\x0b\x32\x15.etcdserverpb.Request\x12)\n\x05range\x18\x03 \x01(\x0b\x32\x1a.etcdserverpb.RangeRequest\x12%\n\x03put\x18\x04 \x01(\x0b\x32\x18.etcdserverpb.PutRequest\x12\x36\n\x0c\x64\x65lete_range\x18\x05 \x01(\x0b\x32 .etcdserverpb.DeleteRangeRequest\x12%\n\x03txn\x18\x06 \x01(\x0b\x32\x18.etcdserverpb.TxnRequest\x12\x33\n\ncompaction\x18\x07 \x01(\x0b\x32\x1f.etcdserverpb.CompactionRequest\x12\x34\n\x0blease_grant\x18\x08 \x01(\x0b\x32\x1f.etcdserverpb.LeaseGrantRequest\x12\x36\n\x0clease_revoke\x18\t \x01(\x0b\x32 .etcdserverpb.LeaseRevokeRequest\x12)\n\x05\x61larm\x18\n \x01(\x0b\x32\x1a.etcdserverpb.AlarmRequest\x12>\n\x10lease_checkpoint\x18\x0b \x01(\x0b\x32$.etcdserverpb.LeaseCheckpointRequest\x12\x35\n\x0b\x61uth_enable\x18\xe8\x07 \x01(\x0b\x32\x1f.etcdserverpb.AuthEnableRequest\x12\x37\n\x0c\x61uth_disable\x18\xf3\x07 \x01(\x0b\x32 .etcdserverpb.AuthDisableRequest\x12\x35\n\x0b\x61uth_status\x18\xf5\x07 \x01(\x0b\x32\x1f.etcdserverpb.AuthStatusRequest\x12@\n\x0c\x61uthenticate\x18\xf4\x07 \x01(\x0b\x32).etcdserverpb.InternalAuthenticateRequest\x12\x38\n\rauth_user_add\x18\xcc\x08 \x01(\x0b\x32 .etcdserverpb.AuthUserAddRequest\x12>\n\x10\x61uth_user_delete\x18\xcd\x08 \x01(\x0b\x32#.etcdserverpb.AuthUserDeleteRequest\x12\x38\n\rauth_user_get\x18\xce\x08 \x01(\x0b\x32 .etcdserverpb.AuthUserGetRequest\x12O\n\x19\x61uth_user_change_password\x18\xcf\x08 \x01(\x0b\x32+.etcdserverpb.AuthUserChangePasswordRequest\x12\x45\n\x14\x61uth_user_grant_role\x18\xd0\x08 \x01(\x0b\x32&.etcdserverpb.AuthUserGrantRoleRequest\x12G\n\x15\x61uth_user_revoke_role\x18\xd1\x08 \x01(\x0b\x32\'.etcdserverpb.AuthUserRevokeRoleRequest\x12:\n\x0e\x61uth_user_list\x18\xd2\x08 \x01(\x0b\x32!.etcdserverpb.AuthUserListRequest\x12:\n\x0e\x61uth_role_list\x18\xd3\x08 \x01(\x0b\x32!.etcdserverpb.AuthRoleListRequest\x12\x38\n\rauth_role_add\x18\xb0\t \x01(\x0b\x32 .etcdserverpb.AuthRoleAddRequest\x12>\n\x10\x61uth_role_delete\x18\xb1\t \x01(\x0b\x32#.etcdserverpb.AuthRoleDeleteRequest\x12\x38\n\rauth_role_get\x18\xb2\t \x01(\x0b\x32 .etcdserverpb.AuthRoleGetRequest\x12Q\n\x1a\x61uth_role_grant_permission\x18\xb3\t \x01(\x0b\x32,.etcdserverpb.AuthRoleGrantPermissionRequest\x12S\n\x1b\x61uth_role_revoke_permission\x18\xb4\t \x01(\x0b\x32-.etcdserverpb.AuthRoleRevokePermissionRequest\x12\x44\n\x13\x63luster_version_set\x18\x94\n \x01(\x0b\x32&.membershippb.ClusterVersionSetRequest\x12K\n\x17\x63luster_member_attr_set\x18\x95\n \x01(\x0b\x32).membershippb.ClusterMemberAttrSetRequest\x12\x42\n\x12\x64owngrade_info_set\x18\x96\n \x01(\x0b\x32%.membershippb.DowngradeInfoSetRequest\"\x0f\n\rEmptyResponse\"S\n\x1bInternalAuthenticateRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\x12\x14\n\x0csimple_token\x18\x03 \x01(\tB\x10\xc8\xe2\x1e\x01\xe0\xe2\x1e\x01\xd0\xe2\x1e\x01\xc8\xe1\x1e\x00\x62\x06proto3')



_REQUESTHEADER = DESCRIPTOR.message_types_by_name['RequestHeader']
_INTERNALRAFTREQUEST = DESCRIPTOR.message_types_by_name['InternalRaftRequest']
_EMPTYRESPONSE = DESCRIPTOR.message_types_by_name['EmptyResponse']
_INTERNALAUTHENTICATEREQUEST = DESCRIPTOR.message_types_by_name['InternalAuthenticateRequest']
RequestHeader = _reflection.GeneratedProtocolMessageType('RequestHeader', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTHEADER,
  '__module__' : 'raft_internal_pb2'
  # @@protoc_insertion_point(class_scope:etcdserverpb.RequestHeader)
  })
_sym_db.RegisterMessage(RequestHeader)

InternalRaftRequest = _reflection.GeneratedProtocolMessageType('InternalRaftRequest', (_message.Message,), {
  'DESCRIPTOR' : _INTERNALRAFTREQUEST,
  '__module__' : 'raft_internal_pb2'
  # @@protoc_insertion_point(class_scope:etcdserverpb.InternalRaftRequest)
  })
_sym_db.RegisterMessage(InternalRaftRequest)

EmptyResponse = _reflection.GeneratedProtocolMessageType('EmptyResponse', (_message.Message,), {
  'DESCRIPTOR' : _EMPTYRESPONSE,
  '__module__' : 'raft_internal_pb2'
  # @@protoc_insertion_point(class_scope:etcdserverpb.EmptyResponse)
  })
_sym_db.RegisterMessage(EmptyResponse)

InternalAuthenticateRequest = _reflection.GeneratedProtocolMessageType('InternalAuthenticateRequest', (_message.Message,), {
  'DESCRIPTOR' : _INTERNALAUTHENTICATEREQUEST,
  '__module__' : 'raft_internal_pb2'
  # @@protoc_insertion_point(class_scope:etcdserverpb.InternalAuthenticateRequest)
  })
_sym_db.RegisterMessage(InternalAuthenticateRequest)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\310\342\036\001\340\342\036\001\320\342\036\001\310\341\036\000'
  _REQUESTHEADER._serialized_start=96
  _REQUESTHEADER._serialized_end=164
  _INTERNALRAFTREQUEST._serialized_start=167
  _INTERNALRAFTREQUEST._serialized_end=2048
  _EMPTYRESPONSE._serialized_start=2050
  _EMPTYRESPONSE._serialized_end=2065
  _INTERNALAUTHENTICATEREQUEST._serialized_start=2067
  _INTERNALAUTHENTICATEREQUEST._serialized_end=2150
# @@protoc_insertion_point(module_scope)



"""
Pure python asyncio Etcd client.
"""

from __future__ import annotations
import asyncio

from typing import (
    AsyncIterator,
    Callable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    OrderedDict,
    Protocol,
    Tuple,
    TypeVar,
    Union,
)
from async_timeout import timeout

import grpc
from grpc.aio import (
    Channel, ClientCallDetails,
    UnaryUnaryCall, UnaryStreamCall,
    StreamUnaryCall, StreamStreamCall,
    UnaryUnaryClientInterceptor, UnaryStreamClientInterceptor,
    StreamUnaryClientInterceptor, StreamStreamClientInterceptor,
    Metadata,
)
from grpc.aio._typing import RequestType, RequestIterableType, ResponseType, ResponseIterableType

from .errors import grpc_exception_handler, match_grpc_error
from .grpc_api import rpc_pb2, rpc_pb2_grpc
from .grpc_api import v3lock_pb2, v3lock_pb2_grpc
from .types import (
    DeleteRangeRequestType, EtcdCredential, EtcdLockOption, HostPortPair,
    PutRequestType, RangeRequestSortOrder, RangeRequestSortTarget, RangeRequestType,
    TransactionRequest, TxnReturnType, TxnReturnValues, WatchCreateRequestFilterType,
    WatchEvent, WatchEventType,
)


__all__ = (
    'EtcdClient',
    'EtcdCommunicator',
    'EtcdConnectionManager',
    'EtcdLockManager',
    'EtcdTransactionAction',
)
T = TypeVar('T', covariant=True)


class Proto(Protocol[T]):
    async def meth(self) -> T:
        pass


class EtcdAuthInterceptor:
    token: str

    def __init__(self, token) -> None:
        self.token = token

    def build_details(self, orig_details: ClientCallDetails) -> ClientCallDetails:
        if _meta := orig_details.metadata:
            _meta.add('token', self.token)
        else:
            _meta = Metadata(('token', self.token))
        return ClientCallDetails(
            orig_details.method,
            orig_details.timeout,
            _meta,
            orig_details.credentials,
            orig_details.wait_for_ready,
        )


class EtcdAuthUnaryUnaryInterceptor(EtcdAuthInterceptor, UnaryUnaryClientInterceptor):
    async def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], UnaryUnaryCall],
        client_call_details: ClientCallDetails,
        request: RequestType,
    ) -> Union[UnaryUnaryCall, ResponseType]:
        return await continuation(
            self.build_details(client_call_details), request)


class EtcdAuthUnaryStreamInterceptor(EtcdAuthInterceptor, UnaryStreamClientInterceptor):
    async def intercept_unary_stream(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], UnaryStreamCall],
        client_call_details: ClientCallDetails,
        request: RequestType,
    ) -> Union[UnaryStreamCall, ResponseIterableType]:
        return await continuation(
            self.build_details(client_call_details), request)


class EtcdAuthStreamUnaryInterceptor(EtcdAuthInterceptor, StreamUnaryClientInterceptor):
    async def intercept_stream_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], StreamUnaryCall],
        client_call_details: ClientCallDetails,
        request_iterator: RequestIterableType,
    ) -> Union[StreamUnaryCall, ResponseType]:
        return await continuation(
            self.build_details(client_call_details), request_iterator)


class EtcdAuthStreamStreamInterceptor(EtcdAuthInterceptor, StreamStreamClientInterceptor):
    async def intercept_stream_stream(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], StreamStreamCall],
        client_call_details: ClientCallDetails,
        request_iterator: RequestIterableType,
    ) -> Union[StreamStreamCall, ResponseIterableType]:
        return await continuation(
            self.build_details(client_call_details), request_iterator)


class EtcdClient:
    """
    Wrapper class of underlying actual Etcd API implementations (KV, Watch, Txn, ...).
    In most cases, user can perform most of the jobs by creating `EtcdClient` object.
    """
    addr: HostPortPair
    _creds: Optional[EtcdCredential]
    secure: bool
    encoding: str

    def __init__(
        self,
        addr: HostPortPair,
        credentials: Optional[EtcdCredential] = None,
        secure: bool = False,
        encoding: str = 'utf-8',
    ) -> None:
        """
        Creates `EtcdClient` instance.

        Parameters
        ---------
        addr
            Connection information of target Etcd cluster.
        credentials
            Authentication information of target Etcd cluster.
            When this value is `None`, `etcetra` will skip all Etcd authentication procedures.
        secure
            If this value is `True`, `etcetra` will try to communicate to Etcd cluster
            with secure gRPC channel.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            Defaults to `utf-8`.
        """
        self.addr = addr
        self._creds = credentials
        self.secure = secure
        self.encoding = encoding

    def _build_channel(self, token: Optional[str] = None):
        chan_cred: Optional[grpc.ChannelCredentials] = None
        if self.secure:
            chan_cred = grpc.ssl_channel_credentials()
        interceptors: Optional[Tuple] = None
        if token is not None:
            interceptors = (
                EtcdAuthUnaryUnaryInterceptor(token),
                EtcdAuthUnaryStreamInterceptor(token),
                EtcdAuthStreamUnaryInterceptor(token),
                EtcdAuthStreamStreamInterceptor(token),
            )

        if chan_cred is not None:
            return grpc.aio.secure_channel(
                str(self.addr), chan_cred, interceptors=interceptors)
        else:
            return grpc.aio.insecure_channel(str(self.addr), interceptors=interceptors)

    def _build_connector_protocol(self) -> Proto[EtcdCommunicator]:
        s = self
        _creds = self._creds

        class P(Proto):
            async def meth(self) -> EtcdCommunicator:
                communicator = EtcdCommunicator(s._build_channel())
                if creds := _creds:
                    token = await communicator._authenticate(creds.username, creds.password)
                    communicator = EtcdCommunicator(s._build_channel(token))
                return communicator
        return P()

    def connect(self) -> EtcdConnectionManager:
        """
        Async context manager which establishes connection to Etcd cluster.

        Returns
        -------
        communicator: EtcdCommunicator
            An `EtcdCommunicator` instance.
        """
        return EtcdConnectionManager(self._build_connector_protocol())

    def with_lock(
        self,
        lock_name: str,
        timeout: Optional[float] = None,
        ttl: Optional[int] = None,
    ) -> EtcdConnectionManager:
        """
        Async context manager which establishes connection and then
        immediately tries to acquire lock with given lock name.
        Acquired lock will automatically released when user exits `with` context.

        Parameters
        ---------
        lock_name
            Name of Etcd lock to acquire.
        timeout
            Number of seconds to wait until lock is acquired. Defaults to `None`.
            If value is `None`, `with_lock` will wait forever until lock is acquired.
        ttl
            If not None, sets a TTL to granted Lock.
            The lock will be automatically released after this amount of seconds elapses.
            Defaults to `None`.
        Returns
        -------
        communicator: EtcdCommunicator
            An `EtcdCommunicator` instance.

        Raises
        -------
        asyncio.TimeoutError
            When timeout expires.
        """
        lock_opt = EtcdLockOption(lock_name, timeout=timeout, ttl=ttl)
        return EtcdConnectionManager(
            self._build_connector_protocol(), encoding=self.encoding, lock_option=lock_opt)


class EtcdConnectionManager:
    communicator_builder: Proto[EtcdCommunicator]
    encoding: str

    _lock_option: Optional[EtcdLockOption]
    _lock: Optional[EtcdLockManager]
    _communicator: Optional[EtcdCommunicator]

    def __init__(
        self,
        communicator_builder: Proto[EtcdCommunicator],
        encoding: str = 'utf-8',
        lock_option: Optional[EtcdLockOption] = None,
    ) -> None:
        self.communicator_builder = communicator_builder
        self.encoding = encoding
        self._lock_option = lock_option
        self._lock = None
        self._communicator = None

    async def __aenter__(self) -> EtcdCommunicator:
        self._communicator = await self.communicator_builder.meth()
        if lock_opt := self._lock_option:
            self._lock = EtcdLockManager(
                lock_opt.lock_name, self._communicator.channel, encoding=self.encoding,
                ttl=lock_opt.ttl, timeout_seconds=lock_opt.timeout)
            await self._lock.__aenter__()
        return self._communicator

    async def __aexit__(self, exc_type, exc, tb) -> Optional[bool]:
        assert self._communicator is not None
        if self._lock is not None:
            await self._lock.__aexit__(exc_type, exc, tb)
        await self._communicator.channel.close()
        return False


class EtcdRequestGenerator:
    @classmethod
    def put(
        cls, key: str, value: Optional[str],
        lease: Optional[int] = None,
        ignore_value: bool = False,
        ignore_lease: bool = False,
        encoding='utf-8',
    ):
        # TODO: Implement prev_kv response
        return rpc_pb2.PutRequest(
            key=key.encode(encoding),
            value=value.encode(encoding) if value else None,
            lease=lease, ignore_lease=ignore_lease, ignore_value=ignore_value,
        )

    @classmethod
    def get(
        cls, key: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        serializable: bool = True,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding='utf-8',
    ):
        return rpc_pb2.RangeRequest(
            key=key.encode(encoding),
            limit=limit,
            max_create_revision=max_create_revision,
            max_mod_revision=max_mod_revision,
            min_create_revision=min_create_revision,
            min_mod_revision=min_mod_revision,
            revision=revision,
            serializable=serializable,
            sort_order=sort_order.value,
            sort_target=sort_target.value,
        )

    @classmethod
    def get_range(
        cls, key: str, range_end: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        serializable: bool = True,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding='utf-8',
    ):
        return rpc_pb2.RangeRequest(
            key=key.encode(encoding),
            range_end=range_end.encode(encoding),
            limit=limit,
            max_create_revision=max_create_revision,
            max_mod_revision=max_mod_revision,
            min_create_revision=min_create_revision,
            min_mod_revision=min_mod_revision,
            revision=revision,
            serializable=serializable,
            sort_order=sort_order.value,
            sort_target=sort_target.value,
        )

    @classmethod
    def get_prefix(
        cls, key: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        serializable: bool = True,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding='utf-8',
    ):
        encoded_key = key.encode(encoding)
        if key[-1] == '/' and len(key) >= 2:
            range_end = encoded_key[:-2] + bytes([encoded_key[-2] + 1]) + b'/'
        else:
            range_end = encoded_key[:-1] + bytes([encoded_key[-1] + 1])
        return rpc_pb2.RangeRequest(
            key=encoded_key,
            range_end=range_end,
            limit=limit,
            max_create_revision=max_create_revision,
            max_mod_revision=max_mod_revision,
            min_create_revision=min_create_revision,
            min_mod_revision=min_mod_revision,
            revision=revision,
            serializable=serializable,
            sort_order=sort_order.value,
            sort_target=sort_target.value,
        )

    @classmethod
    def delete(
        cls, key: str,
        encoding='utf-8',
    ):
        # TODO: Implement prev_kv response
        return rpc_pb2.DeleteRangeRequest(
            key=key.encode(encoding),
        )

    @classmethod
    def delete_range(
        cls, key: str, range_end: str,
        encoding='utf-8',
    ):
        # TODO: Implement prev_kv response
        return rpc_pb2.DeleteRangeRequest(
            key=key.encode(encoding),
            range_end=range_end.encode(encoding),
        )

    @classmethod
    def delete_prefix(
        cls, key: str,
        encoding='utf-8',
    ):
        # TODO: Implement prev_kv response
        encoded_key = key.encode(encoding)
        if key[-1] == '/' and len(key) >= 2:
            range_end = encoded_key[:-2] + bytes([encoded_key[-2] + 1]) + b'/'
        else:
            range_end = encoded_key[:-1] + bytes([encoded_key[-1] + 1])
        return rpc_pb2.DeleteRangeRequest(
            key=encoded_key,
            range_end=range_end,
        )

    @classmethod
    def keys_range(
        cls, key: str, range_end: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        serializable: bool = True,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding='utf-8',
    ):
        return rpc_pb2.RangeRequest(
            key=key.encode(encoding),
            range_end=range_end.encode(encoding),
            keys_only=True,
            limit=limit,
            max_create_revision=max_create_revision,
            max_mod_revision=max_mod_revision,
            min_create_revision=min_create_revision,
            min_mod_revision=min_mod_revision,
            revision=revision,
            serializable=serializable,
            sort_order=sort_order.value,
            sort_target=sort_target.value,
        )

    @classmethod
    def keys_prefix(
        cls, key: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        serializable: bool = True,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding='utf-8',
    ):
        encoded_key = key.encode(encoding)
        if key[-1] == '/' and len(key) >= 2:
            range_end = encoded_key[:-2] + bytes([encoded_key[-2] + 1]) + b'/'
        else:
            range_end = encoded_key[:-1] + bytes([encoded_key[-1] + 1])
        return rpc_pb2.RangeRequest(
            key=encoded_key,
            range_end=range_end,
            keys_only=True,
            limit=limit,
            max_create_revision=max_create_revision,
            max_mod_revision=max_mod_revision,
            min_create_revision=min_create_revision,
            min_mod_revision=min_mod_revision,
            revision=revision,
            serializable=serializable,
            sort_order=sort_order.value,
            sort_target=sort_target.value,
        )


class EtcdCommunicator:
    """
    Performs actual API calls to Etcd cluster and returns result.
    """
    encoding: str
    channel: Channel

    def __init__(
        self,
        channel: Channel,
        encoding: str = 'utf-8',
    ) -> None:
        """
        Creates `EtcdCommunicator` instance.
        In most cases, users won't have to directly create `EtcdCommunicator` class;
        It can be automatically done by `EtcdClient.connect()` or `EtcdClient.with_lock()`.

        Parameters
        ---------
        channel
            Async gRPC Channel to interact.

        encoding
            Character encoding type to encode/decode all types of byte based strings.
            Defaults to `utf-8`.
        """
        self.encoding = encoding
        self.channel = channel

    async def _authenticate(self, username: str, password: str) -> str:
        """
        Tries to authenticate to Etcd server with given credentials.
        In most cases, `EtcdClient` will automatically handle authentication process.
        """
        stub = rpc_pb2_grpc.AuthStub(self.channel)
        response = await stub.Authenticate(
            rpc_pb2.AuthenticateRequest(name=username, password=password),
        )
        return response.token

    @grpc_exception_handler
    async def put(
        self, key: str, value: Optional[str],
        lease: Optional[int] = None,
        prev_kv: bool = False,
        encoding: Optional[str] = None,
    ) -> Optional[str]:
        """
        Puts given key into the key-value store.

        Parameters
        ---------
        key
            The key to put into the key-value store
        value
            The value to associate with the key in the key-value store.
        lease
            The lease ID to associate with the key in the key-value store. Defaults to `None`.
            `None` lease indicates no lease.
        prev_kv
            If this value is `True`, gets the previous value before changing it and returns it.
            Defaults to `False`.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        value: Optional[str]
            If `prev_kv` is set to `True` and previous value exists, returns previous value.
            Otherwise it will just return `None`.
        """
        if encoding is None:
            encoding = self.encoding
        ignore_value = value is None
        if ignore_value:
            value = ''
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.Put(
            EtcdRequestGenerator.put(
                key, value,
                lease=lease, ignore_value=ignore_value,
                encoding=encoding,
            ),
        )
        if prev_kv and response.prev_kv is not None and response.prev_kv.value is not None:
            return response.prev_kv.value.decode(encoding)
        return None

    @grpc_exception_handler
    async def get(
        self, key: str,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        encoding: Optional[str] = None,
    ) -> Optional[str]:
        """
        Gets value associated with given key from the key-value store.

        Parameters
        ---------
        key
            The key to look up.
        max_create_revision
            The upper bound for returned key create revisions;
            all keys with greater create revisions will be filtered away.
        max_mod_revision
            The upper bound for returned key mod revisions;
            all keys with greater mod revisions will be filtered away.
        min_create_revision
            The lower bound for returned key create revisions;
            all keys with lesser create revisions will be filtered away.
        min_mod_revision
            The lower bound for returned key mod revisions;
            all keys with lesser mod revisions will be filtered away.
        revision
            The point-in-time of the key-value store to use for the range.
            If revision is less or equal to zero, the range is over the newest key-value store.
            If the revision has been compacted, ErrCompacted is returned as a response.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        value: Optional[str]
            Returns value if given key exists. Otherwise it will return `None`.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.Range(
            EtcdRequestGenerator.get(
                key,
                max_create_revision=max_create_revision,
                max_mod_revision=max_mod_revision,
                min_create_revision=min_create_revision,
                min_mod_revision=min_mod_revision,
                revision=revision,
                serializable=True,
                sort_order=RangeRequestSortOrder.NONE,
                sort_target=RangeRequestSortTarget.KEY,
                encoding=encoding,
            ),
        )
        if len(response.kvs) > 0:
            return response.kvs[0].value.decode(encoding)
        else:
            return None

    @grpc_exception_handler
    async def get_prefix(
        self, key: str,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding: Optional[str] = None,
    ) -> Mapping[str, str]:
        """
        Gets the key-value in dictionary from the key-value store with given key prefix.
        i.e. `get_prefix('/sorna/local')` call looks up all keys which has `/sorna/local` prefix.

        Parameters
        ---------
        key
            The key prefix to look up.
        max_create_revision
            The upper bound for returned key create revisions;
            all keys with greater create revisions will be filtered away.
        max_mod_revision
            The upper bound for returned key mod revisions;
            all keys with greater mod revisions will be filtered away.
        min_create_revision
            The lower bound for returned key create revisions;
            all keys with lesser create revisions will be filtered away.
        min_mod_revision
            The lower bound for returned key mod revisions;
            all keys with lesser mod revisions will be filtered away.
        revision
            The point-in-time of the key-value store to use for the range.
            If revision is less or equal to zero, the range is over the newest key-value store.
            If the revision has been compacted, ErrCompacted is returned as a response.
        sort_order
            Sort order. Defaults to `RangeRequestSortOrder.NONE`.
        sort_target
            Sort target. Defaults to `RangeRequestSortTarget.KEY`.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        value: Mapping[str, str]
            Returns dictionary with all key-values which matches given key prefix.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.Range(
            EtcdRequestGenerator.get_prefix(
                key,
                max_create_revision=max_create_revision,
                max_mod_revision=max_mod_revision,
                min_create_revision=min_create_revision,
                min_mod_revision=min_mod_revision,
                revision=revision,
                serializable=True,
                sort_order=sort_order,
                sort_target=sort_target,
                encoding=encoding,
            ),
        )
        ret: MutableMapping[str, str] = OrderedDict()
        for x in response.kvs:
            ret[x.key.decode(encoding)] = x.value.decode(encoding)
        return ret

    @grpc_exception_handler
    async def get_range(
        self, key: str, range_end: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding: Optional[str] = None,
    ) -> Mapping[str, str]:
        """
        Gets the key-value in dictionary from the key-value store with keys in [key, range_end) range.

        Parameters
        ---------
        key
            Start of key range.
        range_end
            End of key range.
        max_create_revision
            The upper bound for returned key create revisions;
            all keys with greater create revisions will be filtered away.
        max_mod_revision
            The upper bound for returned key mod revisions;
            all keys with greater mod revisions will be filtered away.
        min_create_revision
            The lower bound for returned key create revisions;
            all keys with lesser create revisions will be filtered away.
        min_mod_revision
            The lower bound for returned key mod revisions;
            all keys with lesser mod revisions will be filtered away.
        revision
            The point-in-time of the key-value store to use for the range.
            If revision is less or equal to zero, the range is over the newest key-value store.
            If the revision has been compacted, ErrCompacted is returned as a response.
        sort_order
            Sort order. Defaults to `RangeRequestSortOrder.NONE`.
        sort_target
            Sort target. Defaults to `RangeRequestSortTarget.KEY`.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        value: Mapping[str, str]
            Returns dictionary with all key-values which matches given key prefix.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.Range(
            EtcdRequestGenerator.get_range(
                key, range_end,
                limit=limit,
                max_create_revision=max_create_revision,
                max_mod_revision=max_mod_revision,
                min_create_revision=min_create_revision,
                min_mod_revision=min_mod_revision,
                revision=revision,
                serializable=True,
                sort_order=sort_order,
                sort_target=sort_target,
                encoding=encoding,
            ),
        )
        ret: MutableMapping[str, str] = OrderedDict()
        for x in response.kvs:
            ret[x.key.decode(encoding)] = x.value.decode(encoding)
        return ret

    @grpc_exception_handler
    async def delete(
        self, key: str,
        prev_kv: bool = False, encoding: Optional[str] = None,
    ) -> Optional[str]:
        """
        Deletes the given key the key-value store.
        A delete request increments the revision of the key-value store
        and generates a delete event in the event history for every deleted key.

        Parameters
        ---------
        key
            The key to delete.
        prev_kv
            If this value set to `True` and previous value with associated target key exists,
            this method will return previous value.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        ------
        value: Optional[str]
            If `prev_kv` is set to `True` and previous value exists, returns previous value.
            Otherwise it will just return `None`.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.DeleteRange(EtcdRequestGenerator.delete(key, encoding=encoding))
        if prev_kv and len(response.prev_kvs) > 0:
            return response.prev_kvs[0].value.decode(encoding)
        else:
            return None

    @grpc_exception_handler
    async def delete_prefix(
        self, key: str,
        prev_kv: bool = False, encoding: Optional[str] = None,
    ) -> Optional[List[Optional[str]]]:
        """
        Deletes keys with given prefix and its associated values from the key-value store.
        A delete request increments the revision of the key-value store
        and generates a delete event in the event history for every deleted key.

        Parameters
        ---------
        key
            The key prefix to delete.
        prev_kv
            If this value set to `True` and previous value with associated target key exists,
            this method will return previous value.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        ------
        values: Optional[List[Optional[str]]]
            If `prev_kv` is set to `True` and previous value exists, returns previous value.
            Otherwise it will just return `None`.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.DeleteRange(EtcdRequestGenerator.delete_prefix(key, encoding=encoding))
        if prev_kv:
            return [
                x.value.decode(encoding) if x.value is not None else None
                for x in response.prev_kvs
            ]
        else:
            return None

    @grpc_exception_handler
    async def delete_range(
        self, key: str, range_end: str,
        prev_kv: bool = False, encoding: Optional[str] = None,
    ) -> Optional[List[Optional[str]]]:
        """
        Deletes the given range from the key-value store.
        A delete request increments the revision of the key-value store
        and generates a delete event in the event history for every deleted key.

        Parameters
        ---------
        key
            Start of key range.
        range_end
            End of key range.
        prev_kv
            If this value set to `True` and previous value with associated target key exists,
            this method will return previous value.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        ------
        values: Optional[str]
            If `prev_kv` is set to `True` and previous value exists, returns previous value.
            Otherwise it will just return `None`.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.DeleteRange(
            EtcdRequestGenerator.delete_range(key, range_end, encoding=encoding))
        if prev_kv:
            return [
                x.value.decode(encoding) if x.value is not None else None
                for x in response.prev_kvs
            ]
        else:
            return None

    @grpc_exception_handler
    async def keys_prefix(
        self, key: str,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding: Optional[str] = None,
    ) -> List[str]:
        """
        Gets the keys which has given prefix from the key-value store.

        Parameters
        ---------
        key
            The key prefix to look up.
        max_create_revision
            The upper bound for returned key create revisions;
            all keys with greater create revisions will be filtered away.
        max_mod_revision
            The upper bound for returned key mod revisions;
            all keys with greater mod revisions will be filtered away.
        min_create_revision
            The lower bound for returned key create revisions;
            all keys with lesser create revisions will be filtered away.
        min_mod_revision
            The lower bound for returned key mod revisions;
            all keys with lesser mod revisions will be filtered away.
        revision
            The point-in-time of the key-value store to use for the range.
            If revision is less or equal to zero, the range is over the newest key-value store.
            If the revision has been compacted, ErrCompacted is returned as a response.
        sort_order
            Sort order. Defaults to `RangeRequestSortOrder.NONE`.
        sort_target
            Sort target. Defaults to `RangeRequestSortTarget.KEY`.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        keys: List[str]
            Returns list of found keys.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.Range(
            EtcdRequestGenerator.keys_prefix(
                key,
                max_create_revision=max_create_revision,
                max_mod_revision=max_mod_revision,
                min_create_revision=min_create_revision,
                min_mod_revision=min_mod_revision,
                revision=revision,
                serializable=True,
                sort_order=sort_order,
                sort_target=sort_target,
                encoding=encoding,
            ),
        )
        return [x.key.decode(encoding) for x in response.kvs]

    @grpc_exception_handler
    async def keys_range(
        self, key: str, range_end: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        serializable: bool = True,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding: Optional[str] = None,
    ) -> List[str]:
        """
        Gets the keys in the range from the key-value store.

        Parameters
        ---------
        key
            Start of key range.
        range_end
            End of key range.
        max_create_revision
            The upper bound for returned key create revisions;
            all keys with greater create revisions will be filtered away.
        max_mod_revision
            The upper bound for returned key mod revisions;
            all keys with greater mod revisions will be filtered away.
        min_create_revision
            The lower bound for returned key create revisions;
            all keys with lesser create revisions will be filtered away.
        min_mod_revision
            The lower bound for returned key mod revisions;
            all keys with lesser mod revisions will be filtered away.
        revision
            The point-in-time of the key-value store to use for the range.
            If revision is less or equal to zero, the range is over the newest key-value store.
            If the revision has been compacted, ErrCompacted is returned as a response.
        sort_order
            Sort order. Defaults to `RangeRequestSortOrder.NONE`.
        sort_target
            Sort target. Defaults to `RangeRequestSortTarget.KEY`.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        keys: List[str]
            Returns list of found keys.
        """
        if encoding is None:
            encoding = self.encoding
        stub = rpc_pb2_grpc.KVStub(self.channel)
        response = await stub.Range(
            EtcdRequestGenerator.keys_range(
                key, range_end,
                limit=limit,
                max_create_revision=max_create_revision,
                max_mod_revision=max_mod_revision,
                min_create_revision=min_create_revision,
                min_mod_revision=min_mod_revision,
                revision=revision,
                serializable=serializable,
                sort_order=sort_order,
                sort_target=sort_target,
                encoding=encoding,
            ),
        )
        return [x.key.decode(encoding) for x in response.kvs]

    @grpc_exception_handler
    async def grant_lease(self, ttl: int, id: Optional[int] = None) -> int:
        """
        Creates a lease which expires if the server does not receive a keepAlive
        within a given time to live period. All keys attached to the lease
        will be expired and deleted if the lease expires.
        Each expired key generates a delete event in the event history.

        Parameters
        ---------
        ttl
            Advisory time-to-live in seconds.
        id
            Requested ID for the lease. If ID is set to None, the lessor chooses an ID.

        Returns
        -------
        id: int
            Lease ID for the granted lease.
        """
        stub = rpc_pb2_grpc.LeaseStub(self.channel)
        response = await stub.LeaseGrant(rpc_pb2.LeaseGrantRequest(ID=id or 0, TTL=ttl))
        return response.ID

    @grpc_exception_handler
    async def revoke_lease(self, id: int) -> None:
        """
        Revokes a lease. All keys attached to the lease will expire and be deleted.

        Parameters
        ---------
        id
            Lease ID to revoke. When the ID is revoked, all associated keys will be deleted.
        """
        stub = rpc_pb2_grpc.LeaseStub(self.channel)
        await stub.LeaseRevoke(rpc_pb2.LeaseRevokeRequest(ID=id))

    def create_lease_keepalive_task(self, id: int, interval: float) -> asyncio.Task:
        """
        Creates asyncio Task which sends Keepalive request to given lease ID.

        Parameters
        ---------
        id
            Lease ID to send Keepalive request.
        interval
            Interval to send Keepalive request.

        Returns
        -------
        task: asyncio.Task
        """

        @grpc_exception_handler
        async def _task_worker():
            stub = rpc_pb2_grpc.LeaseStub(self.channel)

            request = rpc_pb2.LeaseKeepAliveRequest(ID=id)

            stream = stub.LeaseKeepAlive()
            await stream.write(request)

            while True:
                await asyncio.sleep(interval)
                await stream.read()
                request = rpc_pb2.LeaseKeepAliveRequest(ID=id)
                await stream.write(request)

        return asyncio.create_task(_task_worker())

    async def _watch_impl(
        self, key: bytes, encoding: str,
        ready_event: Optional[asyncio.Event] = None,
        filters: Optional[List[WatchCreateRequestFilterType]] = None,
        fragment: bool = False,
        prev_kv: bool = False,
        progress_notify: bool = True,
        range_end: Optional[bytes] = None,
        start_revision: Optional[int] = None,
        watch_id: Optional[int] = None,
    ) -> AsyncIterator[WatchEvent]:
        """
        Actual implementation of `watch` procedure.
        """
        stub = rpc_pb2_grpc.WatchStub(self.channel)

        request = rpc_pb2.WatchRequest()
        request.create_request.key = key
        request.create_request.fragment = fragment
        for e in (filters or []):
            request.create_request.filters.extend([e.value])
        request.create_request.prev_kv = prev_kv
        request.create_request.progress_notify = progress_notify
        if range_end is not None:
            request.create_request.range_end = range_end
        if start_revision is not None:
            request.create_request.start_revision = str(start_revision)
        if watch_id is not None:
            request.create_request.watch_id = str(watch_id)

        stream = stub.Watch()

        try:
            await stream.write(request)
            if ready_event is not None:
                ready_event.set()
            while True:
                response = await stream.read()
                watch_id = response.watch_id
                for event in response.events:
                    if event.type == 0:
                        event_type = WatchEventType.PUT
                    if event.type == 1:
                        event_type = WatchEventType.DELETE
                    if prev_kv and event.prev_kv is not None and event.prev_kv.value is not None:
                        prev_value = event.prev_kv.value.decode(encoding)
                    else:
                        prev_value = None
                    yield WatchEvent(
                        event.kv.key.decode(encoding),
                        event.kv.value.decode(encoding) if event.kv.value is not None else None,
                        prev_value,
                        event_type,
                    )
        except grpc.aio.AioRpcError as e:
            raise match_grpc_error(e) from e
        finally:
            if watch_id is not None and not stream.done():
                request = rpc_pb2.WatchRequest()
                request.cancel_request.watch_id = watch_id
                await stream.write(request)

    def watch(
        self, key: str,
        ready_event: Optional[asyncio.Event] = None,
        filters: Optional[List[WatchCreateRequestFilterType]] = None,
        prev_kv: bool = False,
        progress_notify: bool = False,
        start_revision: Optional[int] = None,
        watch_id: Optional[int] = None,
        encoding: Optional[str] = None,
    ) -> AsyncIterator[WatchEvent]:
        """
        Async iterator which watches for events happening or that have happened.
        Both input and output are streams; the input stream is for creating and canceling watchers
        and the output stream sends events.
        One watch RPC can watch on multiple key ranges, streaming events for several watches at once.
        The entire event history can be watched starting from the last compaction revision.

        Parameters
        ---------
        key
            The key to watch events.
        ready_event
            If this value is set, `Event.set()` will be called
            when watch is ready to accept events.
        filters
            Events to filter. Defaults to `None`.
            If this list is `None`, `watch` will yield all types of event.
        prev_kv
            If this value is set to `True`, event will be yielded with previous value supplied.
        progress_notify
            progress_notify is set so that the etcd server will periodically send a WatchResponse
            with no events to the new watcher if there are no recent events.
            It is useful when clients wish to recover a disconnected watcher
            starting from a recent known revision.
            The etcd server may decide how often it will send notifications based on current load.
        start_revision
            An optional revision to watch from (inclusive). No start_revision is "now".
        watch_id
            If watch_id is provided and non-zero, it will be assigned to this watcher.
            Since creating a watcher in etcd is not a synchronous operation,
            this can be used ensure that ordering is correct
            when creating multiple watchers on the same stream.
            Creating a watcher with an ID already in use on the stream
            will cause an error to be returned.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        event: AsyncIterator[WatchEvent]
            A `WatchEvent` object containing event information.
        """
        if encoding is None:
            encoding = self.encoding

        return self._watch_impl(
            key.encode(encoding), encoding,
            ready_event=ready_event, filters=filters, prev_kv=prev_kv,
            progress_notify=progress_notify, start_revision=start_revision,
            watch_id=watch_id,
        )

    def watch_prefix(
        self, key: str,
        ready_event: Optional[asyncio.Event] = None,
        filters: Optional[List[WatchCreateRequestFilterType]] = None,
        prev_kv: bool = False,
        progress_notify: bool = True,
        start_revision: Optional[int] = None,
        watch_id: Optional[int] = None,
        encoding: Optional[str] = None,
    ) -> AsyncIterator[WatchEvent]:
        """
        Watches for events happening or that have happened along keys with given prefix.
        Both input and output are streams; the input stream is for creating and canceling watchers
        and the output stream sends events.
        One watch RPC can watch on multiple key ranges, streaming events for several watches at once.
        The entire event history can be watched starting from the last compaction revision.

        Parameters
        ---------
        key
            The key prefix to watch events.
        ready_event
            If this value is set, `Event.set()` will be called
            when watch is ready to accept events.
        filters
            Events to filter. Defaults to `None`.
            If this list is `None`, `watch` will yield all types of event.
        prev_kv
            If this value is set to `True`, event will be yielded with previous value supplied.
        progress_notify
            progress_notify is set so that the etcd server will periodically send a WatchResponse
            with no events to the new watcher if there are no recent events.
            It is useful when clients wish to recover a disconnected watcher
            starting from a recent known revision.
            The etcd server may decide how often it will send notifications based on current load.
        start_revision
            An optional revision to watch from (inclusive). No start_revision is "now".
        watch_id
            If watch_id is provided and non-zero, it will be assigned to this watcher.
            Since creating a watcher in etcd is not a synchronous operation,
            this can be used ensure that ordering is correct
            when creating multiple watchers on the same stream.
            Creating a watcher with an ID already in use on the stream
            will cause an error to be returned.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        event: AsyncIterator[WatchEvent]
            A `WatchEvent` object containing event information.
        """
        if encoding is None:
            encoding = self.encoding

        encoded_key = key.encode(encoding)
        if key[-1] == '/' and len(key) >= 2:
            range_end = encoded_key[:-2] + bytes([encoded_key[-2] + 1]) + b'/'
        else:
            range_end = encoded_key[:-1] + bytes([encoded_key[-1] + 1])
        return self._watch_impl(
            key.encode(encoding), encoding,
            ready_event=ready_event, filters=filters, prev_kv=prev_kv,
            range_end=range_end, progress_notify=progress_notify,
            start_revision=start_revision, watch_id=watch_id,
        )

    @grpc_exception_handler
    async def txn(
        self,
        txn_builder: Callable[[EtcdTransactionAction], None],
        encoding: Optional[str] = None,
    ) -> TxnReturnValues:
        """
        A shorthand helper for `Txn`, with no `compare` arguments.
        This can be helpful when user just wants to execute transaction without
        any conditions.

        .. code-block:: python

            >>> await communicator.put('/tmp/successkey', '1111')
            >>> def _txn_builder(action):
            ...     action.get('/tmp/successkey')
            ...
            >>> values = await communicator.txn(_txn_builder)
            >>> print(values)  # ['1111']

        Parameters
        ---------
        txn_builder
            Function which accepts `EtcdTransactionAction` as argument and performs
            all KV calls.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        values: List[TxnReturnType]
            Values returned in each calls inside transaction.
            If the call is `put` or `delete`, `None` will take that place.
        """
        results, _ = await self.txn_compare(
            [],
            lambda success, _: txn_builder(success),
            encoding=encoding,
        )
        return results

    @grpc_exception_handler
    async def txn_compare(
        self,
        compares: List[rpc_pb2.Compare],  # type: ignore
        txn_builder: Callable[[EtcdTransactionAction, EtcdTransactionAction], None],
        encoding: Optional[str] = None,
    ) -> TxnReturnType:
        """
        Processes multiple requests in a single transaction.
        A txn request increments the revision of the key-value store
        and generates events with the same revision for every completed request.
        It is not allowed to modify the same key several times within one txn.

        .. code-block:: python

            >>> from etcetra import CompareKey
            >>> await communicator.put('/tmp/successkey', '1111')
            >>> await communicator.put('/tmp/comparekey', 'asd')
            >>> await communicator.put('/tmp/comparekey2', 'asdg')
            >>> def _txn_builder(success, failure):
            ...     success.get('/tmp/successkey')
            ...
            >>> values = await communicator.txn_compare(
                    [
                        CompareKey('/tmp/comparekey').value == 'asd'],
                        CompareKey('/tmp/comparekey2').value > 'asdf'
                    ],
                    _txn_builder,
                )
            >>> print(values)  # ['1111']

        Parameters
        ---------
        compare
            List of predicates representing a conjunction of terms.
            If the comparisons succeed, then the success requests will be processed in order,
            and the response will contain their respective responses in order.
            If the comparisons fail, then the failure requests will be processed in order,
            and the response will contain their respective responses in order.
        txn_builder
            Function which accepts `EtcdTransactionAction` as argument and performs
            all KV calls.
        encoding
            Character encoding type to encode/decode all types of byte based strings.
            If this value is `None`, this method will use default encoding which is set when creating
            this instance.
            Defaults to `utf-8`.

        Returns
        -------
        values: List[TxnReturnType]
            Values returned in each calls inside transaction.
            If the call is `put` or `delete`, `None` will take that place.
        """
        if encoding is None:
            encoding = self.encoding
        txn = EtcdTransaction(self.channel, encoding=self.encoding)
        txn_builder(txn.success, txn.failure)
        return await txn.execute(compares)


class EtcdTransaction:

    channel: Channel
    encoding: str

    success: EtcdTransactionAction
    failure: EtcdTransactionAction

    def __init__(self, channel: Channel, encoding: str = 'utf-8') -> None:
        self.encoding = encoding
        self.channel = channel

        self.success = EtcdTransactionAction(encoding=encoding)
        self.failure = EtcdTransactionAction(encoding=encoding)

    async def execute(
        self,
        compares: List[rpc_pb2.Compare],  # type: ignore
        encoding: Optional[str] = None,
    ) -> TxnReturnType:
        """
        Executes Txn and returns results.
        """
        if encoding is None:
            encoding = self.encoding
        txn_request = rpc_pb2.TxnRequest()
        txn_request.compare.extend(compares)
        for key in ('success', 'failure'):
            requests: List[TransactionRequest] = getattr(self, key).requests
            for request in requests:
                rop = rpc_pb2.RequestOp()
                if isinstance(request, PutRequestType):
                    rop.request_put.CopyFrom(request)
                elif isinstance(request, RangeRequestType):
                    rop.request_range.CopyFrom(request)
                elif isinstance(request, DeleteRangeRequestType):
                    rop.request_delete_range.CopyFrom(request)
                getattr(txn_request, key).extend([rop])
        stub = rpc_pb2_grpc.KVStub(self.channel)
        result = await stub.Txn(txn_request)

        ret: TxnReturnValues = []
        for response in result.responses:
            response_type = response.WhichOneof('response')
            if response_type == 'response_put':
                ret.append({
                    "revision": response.response_put.header.revision,
                })
            elif response_type == 'response_range':
                ret.append({
                    x.key.decode(encoding): x.value.decode(encoding)
                    for x in response.response_range.kvs
                })
            elif response_type == 'response_delete_range':
                ret.append(None)  # TODO: Handle delete response
            else:
                ret.append(None)
        return TxnReturnType(ret, result.succeeded)


class EtcdTransactionAction:
    """
    Manages calls inside single transaction. `put`, `get` and `delete` calls are supported.
    """
    requests: List[TransactionRequest]
    encoding: str

    callback: Optional[Callable[[bool], None]] = None

    def __init__(self, encoding: str = 'utf-8') -> None:
        self.requests = []
        self.encoding = encoding

    def add_callback(self, cb: Optional[Callable[[bool], None]]) -> None:
        self.callback = cb

    def put(
        self, key: str, value: Optional[str],
        lease: Optional[int] = None,
        ignore_value: bool = False,
        ignore_lease: bool = False,
        encoding: Optional[str] = None,
    ) -> None:
        """
        Puts given key into the key-value store.
        """
        if encoding is None:
            encoding = self.encoding
        self.requests.append(
            EtcdRequestGenerator.put(
                key, value,
                lease=lease, ignore_lease=ignore_lease,
                ignore_value=ignore_value, encoding=encoding,
            ),
        )

    def get(
        self, key: str,
        limit: Optional[str] = None,
        max_create_revision: Optional[str] = None,
        max_mod_revision: Optional[str] = None,
        min_create_revision: Optional[str] = None,
        min_mod_revision: Optional[str] = None,
        revision: Optional[str] = None,
        serializable: bool = True,
        sort_order: RangeRequestSortOrder = RangeRequestSortOrder.NONE,
        sort_target: RangeRequestSortTarget = RangeRequestSortTarget.KEY,
        encoding: Optional[str] = None,
    ) -> None:
        """
        Gets the keys in the range from the key-value store.
        """
        if encoding is None:
            encoding = self.encoding
        self.requests.append(
            EtcdRequestGenerator.get(
                key,
                limit=limit,
                max_create_revision=max_create_revision,
                max_mod_revision=max_mod_revision,
                min_create_revision=min_create_revision,
                min_mod_revision=min_mod_revision,
                revision=revision,
                serializable=serializable,
                sort_order=sort_order,
                sort_target=sort_target,
                encoding=encoding,
            ),
        )

    def delete(self, key: str, encoding: Optional[str] = None) -> None:
        """
        Deletes the given range from the key-value store.
        A delete request increments the revision of the key-value store
        and generates a delete event in the event history for every deleted key.
        """
        if encoding is None:
            encoding = self.encoding
        self.requests.append(EtcdRequestGenerator.delete(key, encoding=encoding))


class EtcdLockManager:
    name: str
    channel: Channel
    encoding: str
    ttl: Optional[int]
    timeout_seconds: Optional[float]

    _lease_id: Optional[int]
    _lock_id: Optional[str]
    _keepalive_task: Optional[asyncio.Task]

    def __init__(
        self,
        name: str,
        channel: Channel,
        encoding: str = 'utf-8',
        ttl: Optional[int] = None,
        timeout_seconds: Optional[float] = None,
    ) -> None:
        self.name = name
        self.channel = channel
        self.encoding = encoding
        self.ttl = ttl
        self.timeout_seconds = timeout_seconds

        self._lease_id = None
        self._lock_id = None
        self._keepalive_task = None

    async def __aenter__(self) -> None:
        """
        Acquires a distributed shared lock on a given named lock.
        On success, it will return a unique key that exists so long as the lock is held by the caller.
        This key can be used in conjunction with transactions to safely ensure updates to etcd
        only occur while holding lock ownership.
        The lock is held until Unlock is called on the key or the lease associate with the owner expires.
        In normal cases `EtcdClient.with_lock()` will automatically handle lock/unlock process.
        """
        stub = v3lock_pb2_grpc.LockStub(self.channel)
        if self.ttl is not None:
            communicator = EtcdCommunicator(self.channel, encoding=self.encoding)
            _lease_id = await communicator.grant_lease(self.ttl)
            self._lease_id = _lease_id
            self._keepalive_task = communicator.create_lease_keepalive_task(
                _lease_id,
                self.ttl / 10,
            )
        else:
            self._lease_id = None
        try:
            async with timeout(self.timeout_seconds):
                response = await stub.Lock(
                    v3lock_pb2.LockRequest(
                        name=self.name.encode(self.encoding),
                        lease=self._lease_id,
                    ),
                )
            self._lock_id = response.key.decode(self.encoding)
        except asyncio.TimeoutError:
            if self._lease_id is not None:
                try:
                    await communicator.revoke_lease(self._lease_id)
                except grpc.aio.AioRpcError as e:
                    if e.code() != grpc.StatusCode.NOT_FOUND:
                        raise e
            raise
        finally:
            if self._keepalive_task is not None:
                self._keepalive_task.cancel()
                try:
                    await self._keepalive_task
                except asyncio.CancelledError:
                    pass

    async def __aexit__(self, exc_type, exc, tb) -> Optional[bool]:
        """
        Releases the hold on lock.
        The next Lock caller waiting for the lock will then be woken up and given ownership of the lock.
        In normal cases `EtcdClient.with_lock()` will automatically handle lock/unlock process.
        """
        assert self._lock_id is not None

        if self._keepalive_task is not None:
            self._keepalive_task.cancel()
            try:
                await self._keepalive_task
            except asyncio.CancelledError:
                pass
        if self._lease_id is not None:
            communicator = EtcdCommunicator(self.channel, encoding=self.encoding)
            try:
                await communicator.revoke_lease(self._lease_id)
            except grpc.aio.AioRpcError as e:
                if e.code() != grpc.StatusCode.NOT_FOUND:
                    raise e
        else:
            stub = v3lock_pb2_grpc.LockStub(self.channel)
            await stub.Unlock(
                v3lock_pb2.UnlockRequest(
                    key=self._lock_id.encode(self.encoding),
                ),
            )
        self._lock_id = None
        self._lease_id = None
        return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v3lock.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from etcetra.grpc_api import gogo_pb2 as gogo__pb2
from etcetra.grpc_api import rpc_pb2 as rpc__pb2
from etcetra.grpc_api import annotations_pb2 as annotations__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cv3lock.proto\x12\x08v3lockpb\x1a\ngogo.proto\x1a\trpc.proto\x1a\x11\x61nnotations.proto\"*\n\x0bLockRequest\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12\r\n\x05lease\x18\x02 \x01(\x03\"I\n\x0cLockResponse\x12,\n\x06header\x18\x01 \x01(\x0b\x32\x1c.etcdserverpb.ResponseHeader\x12\x0b\n\x03key\x18\x02 \x01(\x0c\"\x1c\n\rUnlockRequest\x12\x0b\n\x03key\x18\x01 \x01(\x0c\">\n\x0eUnlockResponse\x12,\n\x06header\x18\x01 \x01(\x0b\x32\x1c.etcdserverpb.ResponseHeader2\xb0\x01\n\x04Lock\x12O\n\x04Lock\x12\x15.v3lockpb.LockRequest\x1a\x16.v3lockpb.LockResponse\"\x18\x82\xd3\xe4\x93\x02\x12\"\r/v3/lock/lock:\x01*\x12W\n\x06Unlock\x12\x17.v3lockpb.UnlockRequest\x1a\x18.v3lockpb.UnlockResponse\"\x1a\x82\xd3\xe4\x93\x02\x14\"\x0f/v3/lock/unlock:\x01*B\x08\xc8\xe2\x1e\x01\xd0\xe2\x1e\x01\x62\x06proto3')



_LOCKREQUEST = DESCRIPTOR.message_types_by_name['LockRequest']
_LOCKRESPONSE = DESCRIPTOR.message_types_by_name['LockResponse']
_UNLOCKREQUEST = DESCRIPTOR.message_types_by_name['UnlockRequest']
_UNLOCKRESPONSE = DESCRIPTOR.message_types_by_name['UnlockResponse']
LockRequest = _reflection.GeneratedProtocolMessageType('LockRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOCKREQUEST,
  '__module__' : 'v3lock_pb2'
  # @@protoc_insertion_point(class_scope:v3lockpb.LockRequest)
  })
_sym_db.RegisterMessage(LockRequest)

LockResponse = _reflection.GeneratedProtocolMessageType('LockResponse', (_message.Message,), {
  'DESCRIPTOR' : _LOCKRESPONSE,
  '__module__' : 'v3lock_pb2'
  # @@protoc_insertion_point(class_scope:v3lockpb.LockResponse)
  })
_sym_db.RegisterMessage(LockResponse)

UnlockRequest = _reflection.GeneratedProtocolMessageType('UnlockRequest', (_message.Message,), {
  'DESCRIPTOR' : _UNLOCKREQUEST,
  '__module__' : 'v3lock_pb2'
  # @@protoc_insertion_point(class_scope:v3lockpb.UnlockRequest)
  })
_sym_db.RegisterMessage(UnlockRequest)

UnlockResponse = _reflection.GeneratedProtocolMessageType('UnlockResponse', (_message.Message,), {
  'DESCRIPTOR' : _UNLOCKRESPONSE,
  '__module__' : 'v3lock_pb2'
  # @@protoc_insertion_point(class_scope:v3lockpb.UnlockResponse)
  })
_sym_db.RegisterMessage(UnlockResponse)

_LOCK = DESCRIPTOR.services_by_name['Lock']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\310\342\036\001\320\342\036\001'
  _LOCK.methods_by_name['Lock']._options = None
  _LOCK.methods_by_name['Lock']._serialized_options = b'\202\323\344\223\002\022\"\r/v3/lock/lock:\001*'
  _LOCK.methods_by_name['Unlock']._options = None
  _LOCK.methods_by_name['Unlock']._serialized_options = b'\202\323\344\223\002\024\"\017/v3/lock/unlock:\001*'
  _LOCKREQUEST._serialized_start=68
  _LOCKREQUEST._serialized_end=110
  _LOCKRESPONSE._serialized_start=112
  _LOCKRESPONSE._serialized_end=185
  _UNLOCKREQUEST._serialized_start=187
  _UNLOCKREQUEST._serialized_end=215
  _UNLOCKRESPONSE._serialized_start=217
  _UNLOCKRESPONSE._serialized_end=279
  _LOCK._serialized_start=282
  _LOCK._serialized_end=458
# @@protoc_insertion_point(module_scope)



import pytest

from etcetra import EtcdClient, EtcdTransactionAction, CompareKey


@pytest.mark.asyncio
async def test_txn(etcd: EtcdClient):
    async with etcd.connect() as communicator:
        await communicator.put('/test/foo', 'bar')
        await communicator.put('/test/asd', '123')

        def _txn(action: EtcdTransactionAction):
            action.get('/test/asd')
            action.put('/test/qwe', 'rty')
            action.get('/test/foo')

        results = await communicator.txn(_txn)
        assert results[0] == {'/test/asd': '123'}
        assert results[1] is not None and results[1].get("revision") is not None
        assert results[2] == {'/test/foo': 'bar'}

        assert (await communicator.get('/test/qwe')) == 'rty'


@pytest.mark.asyncio
async def test_txn_compare_success(etcd: EtcdClient):
    async with etcd.connect() as communicator:
        await communicator.put('/test/cmpkey', 'asdf')

        def _txn(success: EtcdTransactionAction, failure: EtcdTransactionAction):
            success.put('/test/successkey', '123')
            failure.put('/test/failurekey', '456')

        await communicator.txn_compare([
            CompareKey('/test/cmpkey').value == 'asdf',
        ], _txn)

        assert (await communicator.get('/test/successkey')) == '123'
        assert (await communicator.get('/test/failurekey')) is None


@pytest.mark.asyncio
async def test_txn_compare_failure(etcd: EtcdClient):
    async with etcd.connect() as communicator:
        await communicator.put('/test/cmpkey', 'asdf')

        def _txn(success: EtcdTransactionAction, failure: EtcdTransactionAction):
            success.put('/test/successkey', '123')
            failure.put('/test/failurekey', '456')

        await communicator.txn_compare([
            CompareKey('/test/cmpkey').value == 'asdf1234',
        ], _txn)

        assert (await communicator.get('/test/successkey')) is None
        assert (await communicator.get('/test/failurekey')) == '456'



import pytest

from etcetra import EtcdClient


@pytest.mark.asyncio
async def test_put(etcd: EtcdClient):
    async with etcd.connect() as communicator:
        await communicator.put('/test/foo', 'bar')
        assert (await communicator.get('/test/foo')) == 'bar'

        await communicator.put('/test/foo', 'baz')
        assert (await communicator.get('/test/foo')) == 'baz'


@pytest.mark.asyncio
async def test_get(etcd: EtcdClient):
    async with etcd.connect() as communicator:
        await communicator.put('/test/foo', 'bar')
        await communicator.put('/test/foo/bar', 'asd')
        await communicator.put('/test/foo/baz', '1234')
        await communicator.put('/test/foo/baz/qwe', 'rty')

        assert (await communicator.get('/test/foo')) == 'bar'
        assert (await communicator.get('/test/somewhatnonexistingkey')) is None
        assert (await communicator.get_prefix('/test/foo/baz')) == {
            '/test/foo/baz': '1234',
            '/test/foo/baz/qwe': 'rty',
        }
        assert (await communicator.get_prefix('/test/foo/baz/')) == {
            '/test/foo/baz/qwe': 'rty',
        }

        assert (await communicator.keys_prefix('/test/foo')) == [
            '/test/foo', '/test/foo/bar', '/test/foo/baz', '/test/foo/baz/qwe',
        ]

        await communicator.delete('/test/foo/baz/qwe')

        assert (await communicator.keys_prefix('/test/foo')) == [
            '/test/foo', '/test/foo/bar', '/test/foo/baz',
        ]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from etcetra.grpc_api import v3lock_pb2 as v3lock__pb2


class LockStub(object):
    """The lock service exposes client-side locking facilities as a gRPC interface.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Lock = channel.unary_unary(
                '/v3lockpb.Lock/Lock',
                request_serializer=v3lock__pb2.LockRequest.SerializeToString,
                response_deserializer=v3lock__pb2.LockResponse.FromString,
                )
        self.Unlock = channel.unary_unary(
                '/v3lockpb.Lock/Unlock',
                request_serializer=v3lock__pb2.UnlockRequest.SerializeToString,
                response_deserializer=v3lock__pb2.UnlockResponse.FromString,
                )


class LockServicer(object):
    """The lock service exposes client-side locking facilities as a gRPC interface.
    """

    def Lock(self, request, context):
        """Lock acquires a distributed shared lock on a given named lock.
        On success, it will return a unique key that exists so long as the
        lock is held by the caller. This key can be used in conjunction with
        transactions to safely ensure updates to etcd only occur while holding
        lock ownership. The lock is held until Unlock is called on the key or the
        lease associate with the owner expires.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Unlock(self, request, context):
        """Unlock takes a key returned by Lock and releases the hold on lock. The
        next Lock caller waiting for the lock will then be woken up and given
        ownership of the lock.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LockServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Lock': grpc.unary_unary_rpc_method_handler(
                    servicer.Lock,
                    request_deserializer=v3lock__pb2.LockRequest.FromString,
                    response_serializer=v3lock__pb2.LockResponse.SerializeToString,
            ),
            'Unlock': grpc.unary_unary_rpc_method_handler(
                    servicer.Unlock,
                    request_deserializer=v3lock__pb2.UnlockRequest.FromString,
                    response_serializer=v3lock__pb2.UnlockResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'v3lockpb.Lock', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Lock(object):
    """The lock service exposes client-side locking facilities as a gRPC interface.
    """

    @staticmethod
    def Lock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v3lockpb.Lock/Lock',
            v3lock__pb2.LockRequest.SerializeToString,
            v3lock__pb2.LockResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Unlock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/v3lockpb.Lock/Unlock',
            v3lock__pb2.UnlockRequest.SerializeToString,
            v3lock__pb2.UnlockResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: snap.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from etcetra.grpc_api import gogo_pb2 as gogo__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nsnap.proto\x12\x06snappb\x1a\ngogo.proto\"+\n\x08snapshot\x12\x11\n\x03\x63rc\x18\x01 \x01(\rB\x04\xc8\xde\x1f\x00\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x42\x10\xc8\xe2\x1e\x01\xe0\xe2\x1e\x01\xd0\xe2\x1e\x01\xc8\xe1\x1e\x00')



_SNAPSHOT = DESCRIPTOR.message_types_by_name['snapshot']
snapshot = _reflection.GeneratedProtocolMessageType('snapshot', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOT,
  '__module__' : 'snap_pb2'
  # @@protoc_insertion_point(class_scope:snappb.snapshot)
  })
_sym_db.RegisterMessage(snapshot)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\310\342\036\001\340\342\036\001\320\342\036\001\310\341\036\000'
  _SNAPSHOT.fields_by_name['crc']._options = None
  _SNAPSHOT.fields_by_name['crc']._serialized_options = b'\310\336\037\000'
  _SNAPSHOT._serialized_start=34
  _SNAPSHOT._serialized_end=77
# @@protoc_insertion_point(module_scope)



import functools
from typing import Any, Callable, ClassVar, Mapping, Optional

import grpc.aio

from .grpc_api import code_pb2


class EtcetraError(Exception):
    code: ClassVar[int]
    debug_map: Optional[Mapping[str, Any]]

    def __init__(self, message=None, debug_map=None):
        super().__init__(message)
        self.debug_map = debug_map

    def __int__(self):
        return self.code


class EtcdUnknownError(EtcetraError):
    code = code_pb2.UNKNOWN


class EtcdInvalidArgumentError(EtcetraError):
    code = code_pb2.INVALID_ARGUMENT


class EtcdDeadlineExceededError(EtcetraError):
    code = code_pb2.DEADLINE_EXCEEDED


class EtcdNotFoundError(EtcetraError):
    code = code_pb2.NOT_FOUND


class EtcdAlreadyExistsError(EtcetraError):
    code = code_pb2.ALREADY_EXISTS


class EtcdPermissionDeniedError(EtcetraError):
    code = code_pb2.PERMISSION_DENIED


class EtcdUnauthenticatedError(EtcetraError):
    code = code_pb2.UNAUTHENTICATED


class EtcdTooManyRequestError(EtcetraError):
    code = code_pb2.RESOURCE_EXHAUSTED


class EtcdBadRequestError(EtcetraError):
    code = code_pb2.FAILED_PRECONDITION


class EtcdAbortedError(EtcetraError):
    code = code_pb2.ABORTED


class EtcdOutOfRangeError(EtcetraError):
    code = code_pb2.OUT_OF_RANGE


class EtcdUnimplementedError(EtcetraError):
    code = code_pb2.UNIMPLEMENTED


class EtcdInternalError(EtcetraError):
    code = code_pb2.INTERNAL


class EtcdUnavailableError(EtcetraError):
    code = code_pb2.UNAVAILABLE


class EtcdDataLossError(EtcetraError):
    code = code_pb2.DATA_LOSS


def match_grpc_error(e: grpc.aio.AioRpcError):
    match e.code:
        case code_pb2.INVALID_ARGUMENT:
            return EtcdInvalidArgumentError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.DEADLINE_EXCEEDED:
            return EtcdDeadlineExceededError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.NOT_FOUND:
            return EtcdNotFoundError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.ALREADY_EXISTS:
            return EtcdAlreadyExistsError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.PERMISSION_DENIED:
            return EtcdPermissionDeniedError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.UNAUTHENTICATED:
            return EtcdUnauthenticatedError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.RESOURCE_EXHAUSTED:
            return EtcdTooManyRequestError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.FAILED_PRECONDITION:
            return EtcdBadRequestError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.ABORTED:
            return EtcdAbortedError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.OUT_OF_RANGE:
            return EtcdOutOfRangeError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.UNIMPLEMENTED:
            return EtcdUnimplementedError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.INTERNAL:
            return EtcdInternalError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.UNAVAILABLE:
            return EtcdUnavailableError(e.details(), debug_map=e.initial_metadata())
        case code_pb2.DATA_LOSS:
            return EtcdDataLossError(e.details(), debug_map=e.initial_metadata())
    return EtcdUnknownError(e.details(), debug_map=e.initial_metadata())


def grpc_exception_handler(outer: Callable):
    @functools.wraps(outer)
    async def wrapper(*args, **kwargs):
        try:
            return await outer(*args, **kwargs)
        except grpc.aio.AioRpcError as e:
            raise match_grpc_error(e) from e
    return wrapper



import asyncio
from queue import Queue
import time
import traceback
import pytest

from etcetra import EtcdClient


@pytest.mark.asyncio
async def test_lock(etcd: EtcdClient):
    queue: Queue[int] = Queue()

    async def _lock_task_1(queue: Queue[int]):
        global lock_1_entered
        async with etcd.with_lock('/test/locka'):
            queue.put(1)
            await asyncio.sleep(10)

    async def _lock_task_2(queue: Queue[int]):
        global lock_2_entered
        await asyncio.sleep(3)
        async with etcd.with_lock('/test/locka'):
            queue.put(2)
            await asyncio.sleep(10)

    async def _lock_task_3(queue: Queue[int]):
        global lock_3_entered
        await asyncio.sleep(3)
        async with etcd.with_lock('/test/lockb'):
            queue.put(3)
            await asyncio.sleep(10)

    task_1 = asyncio.create_task(_lock_task_1(queue))
    task_2 = asyncio.create_task(_lock_task_2(queue))
    task_3 = asyncio.create_task(_lock_task_3(queue))

    await asyncio.sleep(5)

    assert queue.get() == 1
    assert queue.get() == 3
    assert queue.empty()

    task_1.cancel()
    task_2.cancel()
    task_3.cancel()


@pytest.mark.asyncio
async def test_lock_ttl(etcd: EtcdClient):
    queue: Queue[float] = Queue()

    async def _lock_task_1():
        try:
            async with etcd.with_lock('/test/ttllocka', ttl=3):
                await asyncio.sleep(5)
        except Exception:
            traceback.print_exc()
            raise

    async def _lock_task_2(queue: Queue[float]):
        try:
            start = time.monotonic()
            async with etcd.with_lock('/test/ttllocka', ttl=60):
                queue.put(time.monotonic() - start)
        except Exception:
            traceback.print_exc()
            raise

    task_1 = asyncio.create_task(_lock_task_1())
    await asyncio.sleep(1)
    task_2 = asyncio.create_task(_lock_task_2(queue))
    await asyncio.sleep(5)

    assert not queue.empty()
    assert 1.5 < queue.get() < 2.5
    assert queue.empty()

    task_1.cancel()
    task_2.cancel()


@pytest.mark.asyncio
async def test_lock_ttl_2(etcd: EtcdClient):
    queue: Queue[float] = Queue()

    async def _lock_task_1():
        async with etcd.with_lock('/test/ttl'):
            await asyncio.sleep(3)

    async def _lock_task_2():
        async with etcd.with_lock('/test/ttl', ttl=3):
            await asyncio.sleep(5)

    async def _lock_task_3(queue: Queue[float]):
        start = time.monotonic()
        async with etcd.with_lock('/test/ttl'):
            queue.put(time.monotonic() - start)

    task_1 = asyncio.create_task(_lock_task_1())
    await asyncio.sleep(1)
    task_2 = asyncio.create_task(_lock_task_2())
    await asyncio.sleep(4)
    task_3 = asyncio.create_task(_lock_task_3(queue))
    await asyncio.sleep(3)

    # Timeline
    # T+0: Task 1 starts and acquires the lock
    # T+1: Task 2 starts and tries to acquire the lock
    # T+3: Task 1 releases the lock and task 2 acquires the lock
    # T+5: Task 3 starts and tries to acquire the lock
    # T+6: Task 2 releases the lock and task 3 acquires the lock

    # The test is that TTL starts to count when the lock is acquired (T+3),
    # not when the task tries to acquire the lock (T+1). So 3 seconds TTL
    # expires at T+6, not T+4, blocking the task 3's attempt at T+5.

    assert not queue.empty()
    assert 0.5 < queue.get() < 1.5
    assert queue.empty()

    task_1.cancel()
    task_2.cancel()
    task_3.cancel()



import pytest

from etcetra import EtcdClient, HostPortPair
from etcetra.errors import EtcdUnknownError


@pytest.mark.asyncio
async def test_connect_error():
    etcd = EtcdClient(HostPortPair('127.0.0.1', 1))
    with pytest.raises(EtcdUnknownError):
        async with etcd.connect() as communicator:
            await communicator.get('/this/should/raise/error')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kv.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from etcetra.grpc_api import gogo_pb2 as gogo__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x08kv.proto\x12\x06mvccpb\x1a\ngogo.proto\"u\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\x17\n\x0f\x63reate_revision\x18\x02 \x01(\x03\x12\x14\n\x0cmod_revision\x18\x03 \x01(\x03\x12\x0f\n\x07version\x18\x04 \x01(\x03\x12\r\n\x05value\x18\x05 \x01(\x0c\x12\r\n\x05lease\x18\x06 \x01(\x03\"\x91\x01\n\x05\x45vent\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x17.mvccpb.Event.EventType\x12\x1c\n\x02kv\x18\x02 \x01(\x0b\x32\x10.mvccpb.KeyValue\x12!\n\x07prev_kv\x18\x03 \x01(\x0b\x32\x10.mvccpb.KeyValue\" \n\tEventType\x12\x07\n\x03PUT\x10\x00\x12\n\n\x06\x44\x45LETE\x10\x01\x42\x14\xc8\xe2\x1e\x01\xe0\xe2\x1e\x01\xd0\xe2\x1e\x01\xc8\xe1\x1e\x00\xd0\xe1\x1e\x00\x62\x06proto3')



_KEYVALUE = DESCRIPTOR.message_types_by_name['KeyValue']
_EVENT = DESCRIPTOR.message_types_by_name['Event']
_EVENT_EVENTTYPE = _EVENT.enum_types_by_name['EventType']
KeyValue = _reflection.GeneratedProtocolMessageType('KeyValue', (_message.Message,), {
  'DESCRIPTOR' : _KEYVALUE,
  '__module__' : 'kv_pb2'
  # @@protoc_insertion_point(class_scope:mvccpb.KeyValue)
  })
_sym_db.RegisterMessage(KeyValue)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'kv_pb2'
  # @@protoc_insertion_point(class_scope:mvccpb.Event)
  })
_sym_db.RegisterMessage(Event)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\310\342\036\001\340\342\036\001\320\342\036\001\310\341\036\000\320\341\036\000'
  _KEYVALUE._serialized_start=32
  _KEYVALUE._serialized_end=149
  _EVENT._serialized_start=152
  _EVENT._serialized_end=297
  _EVENT_EVENTTYPE._serialized_start=265
  _EVENT_EVENTTYPE._serialized_end=297
# @@protoc_insertion_point(module_scope)
